// -*- mode: c++; tab: 4 -*-


#ifndef __OBC__
  #include <PalmOS.h>
  #include <DataMgr.h>
  #include <VFSMgr.h>
  #include <StringMgr.h>
#else
  #include "MemHackOBC.h"
#endif

#include "MemHack.h"

#define MAX_FILES	20
#define MAX_STRLEN	255
//#define CROSSOVER
//#define CROSSOVER_TOP
//#define CROSSOVER_LEFT
//#define CROSSOVER_HALF

#define appCreatorID 'Mmhk'
#define appPrefVersion    0

#define PREFERENCE_OPTIONS 1 // Options sizeof(UInt32)
#define PREFERENCE_DBARRAY 2 // dbs I made
#define PREFERENCE_CNARRAY 3 // & their card numbers
#define PREFERENCE_VFSPATH 4 // Path to files - String

// Option bit numbers
#define PREFERENCE_OPTIONS_ENABLED		0
#define PREFERENCE_OPTIONS_ALLOWCARDLAUNCH	1
#define PREFERENCE_OPTIONS_VFS_NOSWITCH 2
#define PREFERENCE_OPTIONS_VFS_NOAUTOSTART	3

#define REG(x) SysNotifyRegister(cn,db,x,NULL,sysNotifyNormalPriority,NULL)
#define UNREG(x) SysNotifyUnregister(cn,db,x,NULL)

static void StartApp(void);
static void EventLoop(void);
static Boolean getOption(short);
static void setOption(short, Boolean);
Boolean vfsSetNotify(void);
void EnableMemHack(void);
static Boolean MainEventHandler(EventType *);
static Boolean appHandleEvent(EventType *);
static void appShutDownHandler(void);
static void PreprocessEvent(EventType *);
void handleVFSField(Boolean);
static char * getVFSPath(void);
static MemHandle getVFSPathHandle(void);
static MemHandle getString(UInt16, UInt16, char *);
static void setString(UInt16, MemHandle);
void quitApp(void);
//void listDrawCallback(Int16, RectangleType, Char **);
void debugBreak(void)  = { 0x4e48 };

static void preparePrefPanel(Boolean, Int16);
Boolean prefLaunchedFromApp;

static void startfn() {
  RectangleType rect;
  rect.topLeft.x = 0;
  rect.topLeft.y = 0;
  rect.extent.y  = 160;
  rect.extent.x  = 160;
  WinEraseRectangle(&rect, 0);
}

static Err updatfn(UInt32 totalBytes, UInt32 offset, void *ignore) {
  RectangleType rect;
  rect.topLeft.x = 0;
  rect.topLeft.y = 78;
  rect.extent.y  = 4;
  rect.extent.x  = ((offset+1)*160L)/(totalBytes+1);
  WinPaintRectangle(&rect, 1);
  return errNone;
}

static void *getControl(UInt16 objID) {
  FormPtr frmP = FrmGetActiveForm();
  return FrmGetObjectPtr(frmP , 
			 FrmGetObjectIndex(frmP , objID));
}

static void dbRemoved(UInt16 crdNo, LocalID dbID) {
  UInt16  done;
  UInt16  total = 0;
  UInt16  i = 0;
  UInt16  prefcn[MAX_FILES];
  LocalID prefdb[MAX_FILES];
  
  i = sizeof(prefcn);
  if (PrefGetAppPreferences(appCreatorID, PREFERENCE_CNARRAY, prefcn,
			    &i, true) == noPreferenceFound || !i)
    return;

  i = sizeof(prefdb);
  if (PrefGetAppPreferences(appCreatorID, PREFERENCE_DBARRAY, prefdb,
			    &i, true) == noPreferenceFound || !i)
    return;

  total = i / sizeof(LocalID);

  for (done = 0; done < total; done++) {
    if ( (prefcn[done] == crdNo) &&
         (prefdb[done] == dbID) ) {
      for (i = done + 1; i < total; i++) {
        prefcn[i - 1] = prefcn[i];
        prefdb[i - 1] = prefdb[i];
      }
      total--;
      PrefSetAppPreferences(appCreatorID, PREFERENCE_DBARRAY, appPrefVersion, 
      			prefdb, total*sizeof(LocalID), true);
      PrefSetAppPreferences(appCreatorID, PREFERENCE_CNARRAY, appPrefVersion, 
      			prefcn, total*sizeof(UInt16),  true);
      break;
    }
  }
}

static void foobar() {
  UInt16  done;
  UInt16  total = 0;
  UInt16  i = 0;
  UInt16  prefcn[MAX_FILES]; 
  LocalID prefdb[MAX_FILES];
  Char *message;

  if (!getOption(PREFERENCE_OPTIONS_ENABLED))
    return;

  i = sizeof(prefcn);
  if (PrefGetAppPreferences(appCreatorID, PREFERENCE_CNARRAY, prefcn,
			    &i, true) == noPreferenceFound || !i)
    return;

  i = sizeof(prefdb);
  if (PrefGetAppPreferences(appCreatorID, PREFERENCE_DBARRAY, prefdb,
			    &i, true) == noPreferenceFound || !i)
    return;

  total = i / sizeof(LocalID);

  startfn();

  message = MemPtrNew(128);
  StrPrintF(message, "Deleting %d Files...", total);
  WinDrawChars(message, StrLen(message), 5, 58);
  MemPtrFree(message);

  for (done = 0; done < total; done++) {
    Err err;
#if 0
    UInt16 props;

    // Dave's suggestion. Crashes all sorts of stuff.
       DmDatabaseInfo(prefcn[iprefdb[i], NULL, &props, NULL, NULL,
		   NULL, NULL, NULL, NULL, NULL, NULL, NULL);

    props &=     props &= ~dmHdrAt    
    DmSetDatabaseInfo(prefcn[i], prefdb[i], NULL, &props, NULL, NULL,
		      NULL, NULL, NULL, NULL, NULL, NULL, NULL);
#endif

    updatfn(total-1, done-1, NULL);

    err = DmDeleteDatabase(prefcn[done], prefdb[done]);
  }  
  startfn();

  PrefSetAppPreferences(appCreatorID, PREFERENCE_DBARRAY, appPrefVersion, NULL, 0, true);
  PrefSetAppPreferences(appCreatorID, PREFERENCE_CNARRAY, appPrefVersion, NULL, 0, true);
}

static void PageInOut(UInt16 cardNo, LocalID dbID) {
  Char *appName = MemPtrNew(MAX_STRLEN + 1);
  Char *dirName = MemPtrNew(MAX_STRLEN + 1);
  Char *fileName = MemPtrNew(MAX_STRLEN + 1);
  Char *vfsPath = NULL;
  UInt32 dirIterator;
  UInt16 volRefNum;
  UInt32 volIterator = vfsIteratorStart;
  Err err;
  short i = 0;
  LocalID dbType;
  UInt16  prefcn[MAX_FILES];
  LocalID prefdb[MAX_FILES];

  if (!getOption(PREFERENCE_OPTIONS_ENABLED))
    goto fail;

  vfsPath = getVFSPath();

  if (!appName || !dirName || !fileName || !vfsPath)
    goto fail;

  err = DmDatabaseInfo (cardNo, dbID, appName, NULL, NULL, NULL, NULL,
			NULL, NULL, NULL, NULL, &dbType, NULL);

  if (err != errNone)
    goto fail;

  if (dbType != sysFileTApplication) 
    goto fail;

  StrPrintF(dirName, vfsPath, appName);

  while (volIterator != vfsIteratorStop) {
    FileInfoType info;
    FileRef dirRef;
    err = VFSVolumeEnumerate(&volRefNum, &volIterator);
    if (err == errNone) {
      // open the directory first, to get the directory reference
      // volRefNum must have already been defined
      err = VFSFileOpen(volRefNum, dirName, vfsModeRead, &dirRef);
      if(err == errNone) {
	info.nameP = fileName; // point to local buffer
	info.nameBufLen = MAX_STRLEN;
	dirIterator = vfsIteratorStart;

	while (dirIterator != vfsIteratorStop) {
	  // Get the next file
	  err = VFSDirEntryEnumerate (dirRef, &dirIterator, &info);
	  if (err == errNone) {
	    if (i >= MAX_FILES)
	      goto fail;

	    startfn();
	    StrPrintF(appName, "Importing %s", fileName);
	    WinDrawChars(appName, StrLen(appName), 5, 58);

	    StrPrintF(appName, 	"%s%s", dirName, fileName);
	    VFSImportDatabaseFromFileCustom(volRefNum, appName, 
					    &prefcn[i], &prefdb[i],
					    updatfn, NULL);

	    startfn();

#if 0
	    // Dave's suggestion. Crashes all sorts of stuff.
	    UInt16 props;
	    DmDatabaseInfo(prefcn[i], prefdb[i], NULL, &props, NULL, NULL,
			   NULL, NULL, NULL, NULL, NULL, NULL, NULL);

	    props |= dmHdrAttrReadOnly;

	    DmSetDatabaseInfo(prefcn[i], prefdb[i], NULL, &props, NULL, NULL,
			      NULL, NULL, NULL, NULL, NULL, NULL, NULL);
#endif

	    i++;
	  } else {
	    goto fail;
	  }
	} /* end while (dirIterator != vfsIteratorStop) */
      } else {
	goto fail;
      }
    } else {
      goto fail;
    }
  } /* end while (volIterator != vfsIteratorStop) */

 fail:
  PrefSetAppPreferences(appCreatorID, PREFERENCE_DBARRAY, appPrefVersion, 
			prefdb, i * sizeof(LocalID), true);
  PrefSetAppPreferences(appCreatorID, PREFERENCE_CNARRAY, appPrefVersion, 
			prefcn, i * sizeof(UInt16),  true);

  if (appName)
    MemPtrFree(appName);
  if (dirName)
    MemPtrFree(dirName);
  if (fileName)
    MemPtrFree(fileName);
  if (vfsPath)
    MemPtrFree(vfsPath);
}

UInt32 PilotMain(UInt16 cmd, MemPtr cmdPBP,
                 UInt16 launchFlags )
{
  switch (cmd) {
     case sysAppLaunchCmdNotify:
       SysNotifyParamType *np = (SysNotifyParamType *)cmdPBP;
       SysNotifyAppLaunchOrQuitType *alp =
         (SysNotifyAppLaunchOrQuitType *)np->notifyDetailsP;
       SysNotifyDBDeletedType *dbd = 
         (SysNotifyDBDeletedType *)np->notifyDetailsP;

//       debugBreak();
       switch(np->notifyType) {
#ifdef CROSSOVER
       case sysNotifyProcessPenStrokeEvent:

         SysNotifyPenStrokeType *psp =
           (SysNotifyPenStrokeType *)np->notifyDetailsP;
         UInt16 numAreas, i;
          const SilkscreenAreaType*Areas = EvtGetSilkscreenAreaList(&numAreas);
         RectangleType  rect;
         for (i=0; i<numAreas; i++)
	          if (Areas[i].areaType == silkscreenRectGraffiti &&
	               Areas[i].index    == alphaGraffitiSilkscreenArea)
	            break;

         if (i == numAreas)
	         break;

#ifdef CROSSOVER_LEFT
         rect.topLeft.x = Areas[i].bounds.topLeft.x;
#else
         rect.topLeft.x = Areas[i].bounds.topLeft.x + 
           (Areas[i].bounds.extent.x >> 1);
#endif /* CROSSOVER_LEFT */
         rect.extent.x  = Areas[i].bounds.extent.x >> 1;

#ifdef CROSSOVER_HALF
         rect.topLeft.y = Areas[i].bounds.topLeft.y;
         rect.extent.y  = Areas[i].bounds.extent.y;
#else
#ifdef CROSSOVER_TOP
         rect.topLeft.y = Areas[i].bounds.topLeft.y;
#else
         rect.topLeft.y = Areas[i].bounds.topLeft.y +
	          (Areas[i].bounds.extent.y >> 1);
#endif /* CROSSOVER_TOP */
         rect.extent.x  = Areas[i].bounds.extent.x >> 1;
         rect.extent.y  = Areas[i].bounds.extent.y >> 1;
#endif /* CROSSOVER_HALF */

         if (RctPtInRectangle(psp->startPt.x, psp->startPt.y, &rect) &&
	             RctPtInRectangle(psp->endPt.x, psp->endPt.y, &rect)) {
	          Boolean cl, nl, as; // caps, num, auto
	          UInt16  sh;
	          GrfGetState(&cl, &nl, &sh, &as);
	          if (!cl && !sh && !as)
	            GrfSetState(cl, nl, 1);
         }

         break;
#endif

       case sysNotifyAppLaunchingEvent:
         PageInOut(alp->cardNo, alp->dbID);
         break;

       case sysNotifyAppQuittingEvent:
         foobar();
         break;

       case sysNotifyDBDeletedEvent:
         dbRemoved(dbd->cardNo, dbd->oldDBID);
         break;

       case sysNotifyVolumeMountedEvent:
         if(getOption(PREFERENCE_OPTIONS_VFS_NOSWITCH)) {
           np->handled |= vfsHandledUIAppSwitch;
         }
         if(getOption(PREFERENCE_OPTIONS_VFS_NOAUTOSTART)) {
           np->handled |= vfsHandledStartPrc;
         }
         break;

     }
     break;

  case sysAppLaunchCmdPanelCalledFromApp:
  case sysAppLaunchCmdNormalLaunch:
  {
    UInt32 dwROMVer;

    FtrGet(sysFtrCreator, sysFtrNumROMVersion, &dwROMVer);
    if (sysGetROMVerMajor(dwROMVer) != 5) {
     FrmAlert(altROM);
      if(sysGetROMVerMajor(dwROMVer) == 1) {
      /* === From: AppLaunchCmd.h === */
/*
#define LaunchWithCommand(type, creator, command, commandParams)  \
{ \
*/
	UInt16			cardNo; 
	LocalID			dbID; 
	DmSearchStateType	searchState; 
	Err 			err; 
	DmGetNextDatabaseByTypeCreator(true, &searchState, sysFileTApplication, 
		sysFileCDefaultApp, true,	&cardNo, &dbID); 
//	ErrNonFatalDisplayIf(!dbID, "Could not find app"); 
	if (dbID) { 
		err = SysUIAppSwitch(cardNo, dbID, sysAppLaunchCmdNormalLaunch, NULL); 
//		ErrNonFatalDisplayIf(err, "Could not launch app"); 
	} 
/*
}} #define AppLaunchWithCommand(appCreator, appCommand, appCommandParams) \
	LaunchWithCommand (sysFileTApplication, appCreator, appCommand, appCommandParams)
	
	AppLaunchWithCommand(sysFileCDefaultApp,
			     sysAppLaunchCmdNormalLaunch, NULL);
*/
      }
      return sysErrRomIncompatible;
    }
    if (cmd == sysAppLaunchCmdNormalLaunch) {
      prefLaunchedFromApp = false;
    } else {
      prefLaunchedFromApp = true;
    }
    StartApp();
 }
    break;

  case sysAppLaunchCmdCardLaunch:
    {
    SysAppLaunchCmdCardType *cl = (SysAppLaunchCmdCardType *)cmdPBP;
    if (!getOption(PREFERENCE_OPTIONS_ALLOWCARDLAUNCH)) {
      // popup a dialog, ask if we should disallow card launch
      if (FrmAlert(altCard) == 0) {
        cl->err = errNone;
      } else {
        setOption(PREFERENCE_OPTIONS_ALLOWCARDLAUNCH, true);
      }
      // in case we got installed as AutoStart...
      cl->startFlags |= sysAppLaunchStartFlagNoAutoDelete;
    }
    }
    break;
  
  case sysAppLaunchCmdSystemReset:
    PrefSetAppPreferences(appCreatorID, PREFERENCE_DBARRAY,  appPrefVersion, NULL, 0, true);
    PrefSetAppPreferences(appCreatorID, PREFERENCE_CNARRAY,  appPrefVersion, NULL, 0, true);
  case sysAppLaunchCmdSyncNotify:
     EnableMemHack();
     break;
  }

  return 0;
}

static void StartApp(void) {
  FrmGotoForm(frmMain);
  EventLoop();
}

static void EventLoop(void) {
  EventType event;
  UInt16 error;

  do {
    EvtGetEvent(&event, evtWaitForever);

    PreprocessEvent(&event);

    if (SysHandleEvent(&event)) continue;
 
    if (MenuHandleEvent(0, &event, &error)) continue;

    if (appHandleEvent(&event)) continue;

    FrmDispatchEvent(&event);

  } while(event.eType != appStopEvent);

  appShutDownHandler();
  FrmCloseAllForms();
}

static Boolean appHandleEvent(EventType *event) {
  switch (event->eType) {

    case frmLoadEvent:
      UInt16 formID = event->data.frmLoad.formID;
      FormPtr pForm = FrmInitForm(formID);
      
      FrmSetActiveForm(pForm);

      switch(formID) {
        case frmMain:
          FrmSetEventHandler(pForm, MainEventHandler);
          return true;
          break;
      }
      break;
        
    case menuEvent:
      switch (event->data.menu.itemID) {
        case mnuAbout:
          FrmHelp(strAbout);
          return true;
          break;

        case mnuHints:
          FrmHelp(strHints);
          return true;
          break;
      }
      break;
  }
  return false;
}

static Boolean MainEventHandler(EventType *event) {
  switch (event->eType) {
    case frmOpenEvent:
      CtlSetValue(getControl(btnEnabled), getOption(PREFERENCE_OPTIONS_ENABLED));
      CtlSetValue(getControl(btnVFSNoSwitch), getOption(PREFERENCE_OPTIONS_VFS_NOSWITCH));
      CtlSetValue(getControl(btnVFSNoAutoStart), getOption(PREFERENCE_OPTIONS_VFS_NOAUTOSTART));

      void * pfld = getControl(fldVFSPath);
      MemHandle txtHandle = FldGetTextHandle(pfld);
      FldSetTextHandle(pfld, getVFSPathHandle());
      if (txtHandle) MemHandleFree(txtHandle);

      EnableMemHack(); // in case of bypass during reset
      preparePrefPanel(true, -2);
      handleVFSField(false);
      FrmDrawForm(FrmGetActiveForm());
      return true;
      break;

    case ctlSelectEvent:
      switch (event->data.ctlEnter.controlID ) {
        case btnEnabled:
          setOption(PREFERENCE_OPTIONS_ENABLED, CtlGetValue(getControl(btnEnabled)));
          EnableMemHack();
          handleVFSField(true);
          return true;
          break;

        case btnVFSPath:
          FrmHelp(strVFSPath);
          return true;
          break; 
        case btnVFSNoSwitch:
          setOption(PREFERENCE_OPTIONS_VFS_NOSWITCH, CtlGetValue(getControl( btnVFSNoSwitch)));
          EnableMemHack();
          return true;
          break; 
        case btnVFSNoAutoStart:
          setOption(PREFERENCE_OPTIONS_VFS_NOAUTOSTART, CtlGetValue(getControl( btnVFSNoAutoStart)));
          EnableMemHack();
          return true;
          break;

        case btnDone:
          quitApp();
          return true;
          break;
      }
      break;
 
    case popSelectEvent:
      Int16 i = event->data.popSelect.selection;
      switch (event->data.popSelect.controlID) {
        case popPanel:
          preparePrefPanel(false, i);
          return true;
          break;
      }
      break;

  }
  return false;
}

static void appShutDownHandler(void) {
  void *    pfld    = getControl(fldVFSPath);
  MemHandle VFSPath = FldGetTextHandle(pfld);
  if (VFSPath) {
    setString(PREFERENCE_VFSPATH, VFSPath);
    FldSetTextHandle(pfld, NULL);
    MemHandleFree(VFSPath);
  }
  preparePrefPanel(false, -2);
}

static void PreprocessEvent(EventType *event) {
  switch (event->eType) {
    case keyDownEvent:
      WChar c = event->data.keyDown.chr;
      if (event->data.keyDown.modifiers &
            commandKeyMask) {
        switch (c) {
          case vchrNavChange:
            if ((event->data.keyDown.keyCode &
                  (navBitsAll | navChangeBitsAll)) !=
                 navChangeSelect)
              break;

          case vchrRockerCenter:
          case vchrThumbWheelPush:
          case vchrJogRelease:
            if (prefLaunchedFromApp) {
          case vchrJogBack:
              quitApp();
            } else {
              EventType e;
              e.eType = ctlSelectEvent;
              e.data.ctlSelect.controlID =
                popPanel;
              e.data.ctlSelect.pControl =
                getControl(popPanel);
              e.data.ctlSelect.on = false;
              e.data.ctlSelect.reserved1 = 0;
              e.data.ctlSelect.value = 0;
              EvtAddEventToQueue(&e);
            }
            break;
        }
      }
      break;
  }
}

static Boolean getOption(short optionNumber) {
  UInt32 ret;
  UInt16 sz;
  sz = sizeof(ret);
  if (PrefGetAppPreferences(appCreatorID, PREFERENCE_OPTIONS, &ret, &sz, 
			    true) == noPreferenceFound) {
    ret = 0;
    PrefSetAppPreferences(appCreatorID, PREFERENCE_OPTIONS, appPrefVersion,
    					&ret, sizeof(ret), true);
  }
  return ( (ret &(UInt32)(((UInt32)1) << optionNumber)) != 0 );
}

static void setOption(short optionNumber, Boolean enabled) {
  UInt32 opt;
  UInt16 sz;
  sz = sizeof(opt);
  if (PrefGetAppPreferences(appCreatorID, PREFERENCE_OPTIONS, &opt, &sz, 
			    true) == noPreferenceFound) {
    opt = 0;
  }
  if (enabled)
    opt |= (UInt32)(((UInt32)1) << optionNumber);
  else
    opt &= ~((UInt32) (((UInt32)1) << optionNumber));

  PrefSetAppPreferences(appCreatorID, PREFERENCE_OPTIONS, appPrefVersion,
  				&opt, sizeof(opt), true);
}


static char * getVFSPath(void) {
  char *    ret  = NULL;
  char *    tmp;
  MemHandle memh = getVFSPathHandle();
  if (memh) {
    tmp = MemHandleLock(memh);
    UInt32 tmplen = StrLen(tmp) + 1;
    ret = MemPtrNew(tmplen);
    if (ret) {
      StrNCopy(ret, tmp, tmplen);
    }
    MemHandleUnlock(memh);
    MemHandleFree(memh);
  }
  return ret;
}

static MemHandle getVFSPathHandle(void) {
  /* get the saved string, or use the default if none exists */
  return getString(PREFERENCE_VFSPATH, MAX_STRLEN, "/Palm/Programs/MemHack/%s/");
}

static MemHandle getString(UInt16 prefNumber, UInt16 maxSize, char *defaultStr) {
  MemHandle ret = MemHandleNew(maxSize + 1);
  char *    ptr;
  UInt16    sz = maxSize;

  if (ret != NULL) {
    ptr = MemHandleLock(ret);
    if (PrefGetAppPreferences(appCreatorID, prefNumber, ptr, &sz,
    				true) == noPreferenceFound) {

      UInt16 bytesToCopy   = StrLen(defaultStr) + 1;
      if (bytesToCopy > maxSize) bytesToCopy = maxSize;
      StrNCopy(ptr, defaultStr, bytesToCopy);
      MemHandleUnlock(ret);
      setString(prefNumber, ret);

    } else {
      MemHandleUnlock(ret);
    }
  }

  return ret;
}

static void setString(UInt16 prefNumber, MemHandle str)  {
  if (str != NULL) {
    char * ptr = MemHandleLock(str);
    UInt16 sz  = StrLen(ptr) + 1;
    PrefSetAppPreferences(appCreatorID, prefNumber, appPrefVersion,
    				ptr, sz, true);
    MemHandleUnlock(str);
  }
}

void EnableMemHack(void) {
  UInt16 cn;
  LocalID  db;
 
  SysCurAppDatabase(&cn, &db);

//  if(getEnabled()) {
  if(getOption(PREFERENCE_OPTIONS_ENABLED)) {
    REG(sysNotifyAppLaunchingEvent);
    REG(sysNotifyAppQuittingEvent);
    REG(sysNotifyDBDeletedEvent);
#ifdef CROSSOVER
    REG(sysNotifyProcessPenStrokeEvent);
#endif
  } else {
    UNREG(sysNotifyAppLaunchingEvent);
    UNREG(sysNotifyAppQuittingEvent);
    UNREG(sysNotifyDBDeletedEvent);
#ifdef CROSSOVER
    UNREG(sysNotifyProcessPenStrokeEvent);
#endif
  }
  if (vfsSetNotify()) {
    REG(sysNotifyVolumeMountedEvent);
  } else {
    UNREG(sysNotifyVolumeMountedEvent);
  }
}

void handleVFSField(Boolean setActive) {
  FormPtr f = FrmGetActiveForm();
  UInt16 i = FrmGetObjectIndex(f, fldVFSPath);
  UInt16 j = FrmGetObjectIndex(f, btnVFSPath);
  if(getOption(PREFERENCE_OPTIONS_ENABLED)) {
    FrmShowObject(f, i);
    FrmShowObject(f, j);
    if (setActive)
      FrmSetFocus(f, i);
  } else {
    FrmHideObject(f, i);
    FrmHideObject(f, j);
    //if (setActive)
    //  FrmSetFocus(f, FrmGetObjectIndex(f, btnEnabled));
  }
}

Boolean vfsSetNotify(void){
SysDBListItemType t;
BitmapPtr m;
FrmGetFormPtr(frmMain);
//SysCreatePanelList(&t, &m);
//MemHandleFree(m);
  if(getOption(PREFERENCE_OPTIONS_VFS_NOSWITCH)) {
    return true;
  }
  if(getOption(PREFERENCE_OPTIONS_VFS_NOAUTOSTART)) {
    return true;
  }
  return false;
}

static void preparePrefPanel(Boolean setup, Int16 popPanelID) {
  FormPtr f = FrmGetActiveForm();
  static MemHandle m = 0, mc = 0;
  static SysDBListItemType *li = NULL;
  static Int16 item = -1;

  if (setup) {
    if (!prefLaunchedFromApp) {
      Char **c = NULL;
      UInt16 n;
      UInt16 lstidx = FrmGetObjectIndex(f, lstPanel);
      ListType *l = FrmGetObjectPtr(f, lstidx);
      FrmHideObject(f, FrmGetObjectIndex(f, btnDone));
      FrmHideObject(f, FrmGetObjectIndex(f, lblPref));
      if (SysCreatePanelList(&n, &m)) {
        if (mc = MemHandleNew(n * sizeof(Char *))) {
          UInt16 i;
          UInt16 cn;
          LocalID  db;
          Int16 maxWidth = 0;
          FontID prevID = FntSetFont(stdFont);

          SysCurAppDatabase(&cn, &db);
          li = MemHandleLock(m);
          c = MemHandleLock(mc);
          for (i = 0; i < n; i++) {
            Int16 w = FntLineWidth(li[i].name,
                                   StrLen(li[i].name));
            c[i] = li[i].name;
            if ((li[i].cardNo == cn) &&
                  (li[i].dbID == db)) {
              item = i;
            }
            if (w > maxWidth)
              maxWidth = w;
          }
          FntSetFont(prevID);
          LstSetListChoices(l, c, (Int16) n);
          LstSetSelection(l, item);
          LstSetHeight(l, n);

          RectangleType b;
          FrmGetObjectBounds(f, lstidx, &b);
          b.extent.x = maxWidth + 4;
          b.topLeft.x = 156 - maxWidth;
          FrmSetObjectBounds(f, lstidx, &b);
        }
      }
    } else {
      FrmHideObject(f, FrmGetObjectIndex(f, popPanel));
    }
  } else { // !setup
    if (popPanelID < -1) {
      if (!prefLaunchedFromApp) {
        if (m) {
          MemHandleUnlock(m);
          MemHandleFree(m);
        }
        if (mc) {
          MemHandleUnlock(mc);
          MemHandleFree(mc);
       }
     }
   } else {
     if ((popPanelID >= 0) &&
           (popPanelID != item)) {
       SysUIAppSwitch(li[popPanelID].cardNo,
                                     li[popPanelID].dbID,
                                     sysAppLaunchCmdNormalLaunch,
                                     NULL);
      }
    }
  } // end if setup
}

void quitApp(void) {
  EventType e;
  e.eType = appStopEvent;
  EvtAddEventToQueue(&e);
}
