#ifndef __MemHackOBC_H__
#define __MemHackOBC_H__

#ifdef __OBC__

/* === From: CoreTraps.h === */

#define sysTrapSysUIAppSwitch			0xa0a7
#define sysTrapSysCurAppDatabase		0xa0ac
#define sysTrapFrmHideObject			0xa184
#define sysTrapFrmShowObject			0xa185
#define sysTrapLstSetDrawFunction		0xa1b0
#define sysTrapLstDrawList			0xa1b1
#define sysTrapLstSetHeight			0xa1b6
#define sysTrapRctPtInRectangle		0xa1c8
#define sysTrapGrfGetState        		0xa280
#define sysTrapGrfSetState        		0xa281
#define sysTrapSysCreatePanelList  		0xa2dc
#define sysTrapFrmSetObjectBounds		0xa303
#define sysTrapFileSystemDispatch		0xa348
#define sysTrapSysNotifyRegister		0xa353
#define sysTrapSysNotifyUnregister		0xa354
#define sysTrapEvtGetSilkscreenAreaList	0xa3cb


/* === From: VFSMgr.h === */

#define VFSMGR_TRAP(x) = { 0x7400 + x, 0x4e4f, sysTrapFileSystemDispatch } 

typedef struct FileInfoTag {
	UInt32	attributes;
	Char		*nameP;	// buffer to receive full name; pass NULL to avoid getting name
	UInt16	nameBufLen; // size of nameP buffer, in bytes
} FileInfoType;
typedef FileInfoType *FileInfoPtr;
typedef UInt32 FileRef;

#define vfsModeRead             (0x0002U)

/************************************************************
 * Iterator start and stop constants.
 * Used by VFSVolumeEnumerate, VFSDirEntryEnumerate, VFSDirEntryEnumerate
 *************************************************************/
#define vfsIteratorStart	0L
#define vfsIteratorStop		0xffffffffL

/************************************************************
 * 'handled' field bit constants 
 * (for use with Volume Mounted/Unmounted notifications)
 *************************************************************/
#define vfsHandledUIAppSwitch	0x01	// Any UI app switching has already been handled.  
						// The VFSMgr will not UIAppSwitch to the start.prc app 
						// (but it will loaded & sent the AutoStart launchcode), 
						// and the Launcher will not switch to itself.
#define vfsHandledStartPrc	0x02	// And automatic running of start.prc has already been handled.
						// VFSMgr will not load it, send it the AutoStart launchcode,
						// or UIAppSwitch to it.

#define vfsTrapFileOpen					3
#define vfsTrapDirEntryEnumerate			20
#define vfsTrapVolumeEnumerate			27
#define vfsTrapImportDatabaseFromFileCustom	39

typedef Err (*VFSImportProcPtr)
	(UInt32 totalBytes, UInt32 offset, void *userDataP);
 
Err VFSFileOpen(UInt16 volRefNum, const Char *pathNameP, UInt16 openMode, FileRef *fileRefP)
	VFSMGR_TRAP(vfsTrapFileOpen);
Err VFSDirEntryEnumerate(FileRef dirRef, UInt32 *dirEntryIteratorP, FileInfoType *infoP)
	VFSMGR_TRAP(vfsTrapDirEntryEnumerate);
Err VFSVolumeEnumerate(UInt16 *volRefNumP, UInt32 *volIteratorP)
	VFSMGR_TRAP(vfsTrapVolumeEnumerate);
Err VFSImportDatabaseFromFileCustom(UInt16 volRefNum, const Char *pathNameP, UInt16 *cardNoP,
		LocalID *dbIDP, VFSImportProcPtr importProcP, void *userDataP)
	VFSMGR_TRAP(vfsTrapImportDatabaseFromFileCustom);


/* === From: SystemResources.h === */

#define	sysFileTApplication			'appl'	// File type for applications


/* === From: ErrorBase.h === */

#define	sysErrorClass				0x0500	// System Manager


/* === From: ErrorMgr.h , with error checking off === */

//#define ErrNonFatalDisplayIf(cond,msg)


/* === From: SystemMgr.h === */

#define sysAppLaunchCmdSyncNotify	3	// Sent to apps whose databases changed during
							// HotSync after the sync has been completed,
							// including when the app itself has been installed
							// by HotSync. The data database(s) must have the
							// same creator ID as the application for this
							// mechanism to function correctly. This is a
							// good opportunity to update/initialize/validate
							// the app's data, such as resorting records,
							// setting alarms, etc.
							//
							// Parameter block: None.
							// Restrictions: No accessing of global or
							//		static variables; no User Interface calls.
							// Notes: This action code is sent via a
							//		direct function call into the app's
							//		PilotMain function from the background
							//		thread of the HotSync application.

#define sysAppLaunchCmdSystemReset	5	// Sent to all applications and preference
							// panels when the system is either soft-reset
							// or hard-reset.  This notification is the
							// right place to ini right place tovalidate
							// your application's preferences/features/
							// database(s) as well as to update alarms and
							// other time-related activities and resources.
							//
							// Parameter block: SysAppLaunchCmdSystemResetType
							// Restrictions: No accessing of global or
							//		static variables; no User Interface calls.
							// Notes: This action code is sent via a direct
							//		function call into the app's PilotMain
							//		function without "launching" the app.

#define sysAppLaunchCmdPanelCalledFromApp 13 // The panel should display a done
							// button instead of the pick list
							// The Done button will return the user
							// to the last app.

#define sysAppLaunchCmdNotify		51	// This is a NotifyMgr notification sent 
							// via SysNotifyBroadcast.  The cmdPBP parameter
							// points to a SysNotifyParamType structure
							// containing more specific information
							// about the notification (e.g., what it's for).

#define sysAppLaunchCmdCardLaunch	58	// Sent to an application by the Launcher when the
							// application is being run from a card.

// For sysAppLaunchCmdCardLaunch
typedef struct {
	Err			err;
	UInt16		volRefNum;
	const Char	*path;
	UInt16		startFlags;			// See vfsStartFlagXXX constants below
} SysAppLaunchCmdCardType;

typedef struct {
	Char			name[dmDBNameLength];
	UInt32	creator;
	UInt32 type;
	UInt16	version;
	LocalID	dbID;
	UInt16	cardNo;
	BitmapPtr	iconP;
} SysDBListItemType;

#define sysAppLaunchStartFlagAutoStart	0x0001	// this bit in the 'startFlags' field is set for an app which is run automatically on card insertion
#define sysAppLaunchStartFlagNoUISwitch	0x0002	// set this bit in the 'startFlags' field to prevent a UI switch to the start.prc app
#define sysAppLaunchStartFlagNoAutoDelete	0x0004	// set this bit in the 'startFlags' field to prevent VFSMgr from deleting start.prc app on volume unmount

#define sysErrRomIncompatible			(sysErrorClass | 12)

#define sysGetROMVerMajor(dwROMVer)		(((UInt16)((dwROMVer) >> 24)) & 0x00FF)

Err SysCurAppDatabase (UInt16 *cardNoP, LocalID *dbIDP)
        SYS_TRAP(sysTrapSysCurAppDatabase);

Err SysUIAppSwitch (UInt16 cardNo, LocalID dbID, UInt16 cmd, MemPtr cmdPBP)
        SYS_TRAP(sysTrapSysUIAppSwitch);

Boolean SysCreatePanelList (UInt16 *panelCount, MemHandle *panelIDs)
        SYS_TRAP(sysTrapSysCreatePanelList);


/* === From: NotifyMgr.h === */

typedef struct SysNotifyParamType {
	UInt32		notifyType;	// What type of event occurred?
	UInt32		broadcaster;	// normally creator code of broadcasting app
	void *		notifyDetailsP;	// ptr to notification-specific data, if any
	void *		userDataP;	// user specified ptr passed back with notification
	Boolean		handled;	// true if event is handled yet
	UInt8		reserved2;
} SysNotifyParamType;

typedef Err  (*SysNotifyProcPtr) (SysNotifyParamType *notifyParamsP);

#define sysNotifyNormalPriority   0     // clients should use this priority

#define sysNotifySleepRequestEvent	'slpq'	// A "Sleep Request Event" is broadcast when the device
								// is about to go to sleep, and is a chance for clients
								// to do perform an action or even delay going to sleep
								// for a little while.  This may be broadcast multiple 
								// times if one or more clients defer the sleep. Parameter 
								// is a pointer to a SleepEventParamType.  See below 
								// details on this structure. Note that this event is 
								// NOT guaranteed to be broadcast.  See the big comment 
								// at the top of this list for important detials & warnings.
								// Always sent from UI thread.

#define sysNotifySleepNotifyEvent	'slp!'	// A "Sleep Notify Event" is broadcast when the device is 
								// definitely going to sleep as soon as the broadcast 
								// is finished.  Parameter is unused.
								// Note that this event is NOT guaranteed to be broadcast.  
								// See the big comment at the top of this list for 
								// important detials & warnings.
								// Always sent from UI thread.

#define sysNotifyEarlyWakeupEvent	'worm'	// ...the early bird gets the worm...
								// Broadcast just after the device wakes up
								// at the early stage where the screen may 
								// still be turned off and we may quickly go 
								// back to sleep after handling an alarm or 
								// charger event.  
								// Always sent from UI thread.

#define sysNotifyLateWakeupEvent	'lazy'	// Broadcast after the device wakes up
								// at the later stage of the wakeup 
								// process after we turn the screen on, 
								// broadcast from EvtResetAutoOffEvent 
								// the first time it is called implying
								// that the device will remain awake 
								// for at least a little while.
								// Always sent from UI thread.

#define sysNotifyProcessPenStrokeEvent	'hpps'	// Sent to enable custom recognition of strokes on the system area of the digitizer (buttons, etc.)
									// Note: future OS releases might not send this notification.

#define sysNotifyAppLaunchingEvent	'hapl'	// Sent before an app is launched (normalLaunch only)
								// Note: future OS releases might not send this notification.

#define sysNotifyAppQuittingEvent	'hapq'	// Sent right after an app has quit (normalLaunch only)
								// Note: future OS releases might not send this notification.

#define sysNotifyVolumeMountedEvent	'volm'	// Broadcast when a VFSMgr volume is 
								// mounted, Always broadcast from UI task.
								// VFSMgr will run start.prc (if present),
								// and SysUIAppSwitch  to it or the Launcher
								// unless the appropriate bits in the 'handled'
								// field are set by a notification handler.
								// PARAMETER: VFSAnyMountParamPtr cast as void*

#define sysNotifyDBDeletedEvent	'dbs-'	// broadcast AFTER a database is removed from the device.
								// Note that the database ID in the parameter block is
								// NO LONGER VALID, and you WILL CRASH the device if you
								// try to pass it to any DataMgr routines.
								// notifyDetailsP: ptr to SysNotifyDBDeletedType.

// for sysNotifyAppLaunchingEvent and sysNotifyAppQuittingEvent
typedef struct SysNotifyAppLaunchOrQuitTag {
	UInt32    version;   // this is version 0
	UInt32    dbID; //  application being started or quitting (note: type unsafeness intentional)
	UInt16    cardNo;    // the application being started or quitting
} SysNotifyAppLaunchOrQuitType;

// for sysNotifyProcessPenStrokeEvent
typedef struct SysNotifyPenStrokeTag {
	UInt32     version;   // this is version 0
	PointType  startPt;
	PointType  endPt;
} SysNotifyPenStrokeType;

// Possible values for the sleep-reason for the sysNotifySleepEvent
#define sysSleepPowerButton		0
#define sysSleepAutoOff			1
#define sysSleepResumed			2
#define sysSleepUnknown			3

// for sysNotifySleepRequestEvent
typedef struct SleepEventParamTag {
	UInt16 reason;		// contains the reason we are going to sleep. See above list for possible values.
	UInt16 deferSleep;	// Only used for the sysNotifySleepRequestEvent, this should be incremented by 
					// clients that want to defer the sleep.  They are then responsible for enqueing
					// a resumeSleepChr in the event queue.
} SleepEventParamType;

// for sysNotifyDBDeletedEvent
typedef struct SysNotifyDBDeletedTag {
	LocalID	oldDBID;		// old database ID of dabatase, no longer valid
	UInt16	cardNo;		// card number that dabatase resides on
	UInt16	attributes;		// database attributes
	Char	dbName[dmDBNameLength];
	UInt32	creator;		// creator code of dabatase
	UInt32	type;			// type code of dabatase
} SysNotifyDBDeletedType;

Err SysNotifyRegister(UInt16 cardNo, LocalID dbID, UInt32 notifyType,
                                         SysNotifyProcPtr callbackP, Int8 priority, void *userDataP)
                         SYS_TRAP(sysTrapSysNotifyRegister);

Err SysNotifyUnregister(UInt16 cardNo, LocalID dbID, UInt32 notifyType, Int8 priority)
                         SYS_TRAP(sysTrapSysNotifyUnregister);


/* === From: SysEvtMgr.h === */

/************************************************************
 * Silkscreen area info structure. An array of these structures
 * is returned by the EvtGetSilkscreenAreaList function.
 *************************************************************/

// Different types of rectangles on the display. For new vendor areas,
// the type should be set to the vendor's creator code, as assigned
// by 3Com's Partnergroup.
#define	silkscreenRectScreen	'scrn'
#define	silkscreenRectGraffiti	'graf'

// Values for SilkscreenAreaType.index if areaType = silkscreenRectGraffiti
#define	alphaGraffitiSilkscreenArea		0
#define	numericGraffitiSilkscreenArea		1

// One silkscreen area. The areaType field tells us which type of
// area it is, while the index field has different meanings depending
// on the area type.
typedef struct SilkscreenAreaType {
	RectangleType			bounds;
	UInt32				areaType;	// four byte creator code.
	UInt16				index;
} SilkscreenAreaType;

// Return pointer to the silkscreen area list
const SilkscreenAreaType* EvtGetSilkscreenAreaList(UInt16* numAreas)
	SYS_TRAP(sysTrapEvtGetSilkscreenAreaList);


/* === From: Rect.h === */

extern Boolean RctPtInRectangle (Coord x, Coord y, const RectangleType *rP)
	SYS_TRAP(sysTrapRctPtInRectangle);


/* === From: Graffiti.h === */

Err GrfGetState(Boolean *capsLockP, Boolean *numLockP, 
		UInt16 *tempShiftP, Boolean *autoShiftedP)
	SYS_TRAP(sysTrapGrfGetState);

Err GrfSetState(Boolean capsLock, Boolean numLock, 
		Boolean upperShift)
	SYS_TRAP(sysTrapGrfSetState);


/* === From: Form.h === */

void FrmHideObject(FormType *formP, UInt16 objIndex)
		SYS_TRAP(sysTrapFrmHideObject);
void FrmShowObject(FormType *formP, UInt16 objIndex)
		SYS_TRAP(sysTrapFrmShowObject);
void FrmSetObjectBounds(FormType *formP, UInt16 objIndex, const RectangleType *bounds)
		SYS_TRAP(sysTrapFrmSetObjectBounds);


/* === From: List.h === */

void LstSetDrawFunction(ListType *listP, ListDrawDataFuncPtr func)
    SYS_TRAP(sysTrapLstSetDrawFunction);
void LstDrawList(ListType *listP)
    SYS_TRAP(sysTrapLstDrawList);
void LstSetHeight(ListType *listP, Int16 visibleItems)
    SYS_TRAP(sysTrapLstSetHeight);


/* === From: SysEvent.h === */

#define commandKeyMask	0x0008


/* === From: Chars.h === */

#define vchrThumbWheelPush	0x0130
#define vchrRockerCenter	0x0136


/* === From: SonyChars.h === */

#define vchrJogPush	0x1705
#define vchrJogRelease	0x1706
#define vchrJogBack	0x1707


/* === From: PalmChars.h === */

#define vchrNavChange		0x0503
#define navBitsAll		0x001f
#define navChangeBitsAll	0x1f00
#define navChangeSelect		0x1000


#endif // __OBC__

#endif // __MemHackOBC_H__
